import glob
import os
import subprocess
import time
from typing import Callable, Optional

from cfn_guard_test.suites import CfnGuardTestSuites
from cfn_guard_test.reader import CfnGuardReader


class CfnGuardRunner:
    """
    Understands how to execute cfn-guard.
    """

    def __init__(
        self,
        cfn_guard_path: str,
        rules_path: str,
        test_path: str,
        output_callback: Optional[Callable[[str], None]] = None,
    ):
        self.cfn_guard_path = cfn_guard_path
        self.rules_path = rules_path.rstrip("/")
        self.test_path = test_path.rstrip("/")
        self.output_callback = output_callback

    def __print(self, message: str) -> None:
        if self.output_callback:
            self.output_callback(message)

    @property
    def __test_cases(self) -> dict:
        rule_sets = {}
        tests = glob.glob(f"{self.test_path}/*_tests.yaml")

        for test in tests:
            test_name = os.path.basename(test)
            rule_name = test_name.split("_")[0]

            if os.path.isfile(f"{self.rules_path}/{rule_name}.guard"):
                rule_sets[test] = f"{self.rules_path}/{rule_name}.guard"

        return rule_sets

    def execute(self) -> CfnGuardTestSuites:
        suite = CfnGuardTestSuites()

        for rules_test_file, rules_file in self.__test_cases.items():
            start_time = time.time()
            self.__print(f"Running {rules_test_file}")
            response = subprocess.run(
                [
                    self.cfn_guard_path,
                    "test",
                    "--rules-file",
                    rules_file,
                    "--test-data",
                    rules_test_file,
                ],
                stdout=subprocess.PIPE,
            )
            end_time = time.time()
            CfnGuardReader(
                suite, rules_file, response.stdout, duration=end_time - start_time
            )

        return suite



# import os
# from click.testing import CliRunner
# from cfn_guard_test import main
#
#
# def test_custom_set() -> None:
#     """
#     Used to load a custom test set in other projects, this is typically commented out and useful when you need to debug!
#     :return:
#     """
#
#     runner = CliRunner()
#     result = runner.invoke(main, ["--rules-path", os.path.expanduser("~/workspace/cfn-guard-cis-benchmark/rules")])
#     assert result.exit_code == 0



from typing import Optional, Callable

import click

from cfn_guard_test.case import CfnGuardTestCase
from cfn_guard_test.report import CfnGuardReport
from cfn_guard_test.rule import CfnGuardRule
from cfn_guard_test.runner import CfnGuardRunner
from cfn_guard_test.suite import CfnGuardTestSuite
from cfn_guard_test.suites import CfnGuardTestSuites
from cfn_guard_test.reader import CfnGuardReader


def validate_cfn_guard_path(
    ctx: click.Context, _: click.Option, value: Optional[str]
) -> str:
    if not value or ctx.resilient_parsing:
        return "cfn-guard"

    if value.split("/")[-1] != "cfn-guard":
        raise click.ClickException(
            "When a custom cfn-guard path is given it needs to end on cfn-guard."
        )

    return value


def resolve_paths(ctx: click.Context, _: click.Option, value: Optional[str]) -> str:
    if not value:
        return ctx.params.get("rules_path", "rules")

    return value


def validate_writable_path(
    ctx: click.Context, _: click.Option, value: Optional[str]
) -> Optional[str]:

    try:
        if value:
            with open(value, "w") as fh:
                if not fh.writable():
                    raise Exception()
    except FileNotFoundError:
        raise click.ClickException(f"The given path is not writable: {value}")
    except Exception:
        raise click.ClickException(f"The given path is not writable: {value}")

    return value


@click.command()
@click.option("-r", "--rules-path", callback=resolve_paths)
@click.option("-t", "--test-path", callback=resolve_paths)
@click.option("--cfn-guard-path", callback=validate_cfn_guard_path)
@click.option("--junit-path", callback=validate_writable_path)
@click.option("-v", "--verbose", count=True)
def main(
    rules_path: str,
    test_path: str,
    cfn_guard_path: str,
    junit_path: Optional[str],
    verbose: int,
) -> None:
    """
    cfn-guard-test

    This tool allows you to easily run your cfn-guard tests against your cfn-guard rules.
    """
    click.echo("===== Analyzing Results =====")
    click.echo()

    suites = CfnGuardRunner(
        cfn_guard_path=cfn_guard_path,
        rules_path=rules_path,
        test_path=test_path,
        output_callback=resolve_output_callback(verbose),
    ).execute()

    if junit_path:
        CfnGuardReport(suites).write(junit_path)

    click.echo()
    click.echo(f"Errors {suites.errors}")
    click.echo(f"Failed {suites.failed}")
    click.echo(f"Passed {suites.passed}")
    click.echo()
    display_failures_errors(suites)


def display_failures_errors(suites: CfnGuardTestSuites) -> None:

    if suites.errors:
        click.echo(click.style("Errors:", bold=True))
        list(map(click.echo, suites.error_messages))
        click.echo()

    if suites.failed:
        click.echo(click.style("Failures:", bold=True))
        list(map(click.echo, suites.failure_messages))
        click.echo()

    if suites.errors or suites.failed:
        exit(1)


def resolve_output_callback(verbose) -> Optional[Callable[[str], None]]:
    return click.echo if verbose else None


if __name__ == "__main__":
    main()



class CfnGuardRule:
    """
    Understands the outcome of a specific rule.

    A single cfn-guard rule contains the result of the evaluated rule. Each rule definition in the `expectations`
    section of the `<name>_tests.yaml` file results in a cfn-guard rule.
    """

    __name: str
    __expected: str
    __evaluated: str

    def __init__(self, name: str, expected: str, evaluated: str) -> None:
        self.__name = name
        self.__expected = expected
        self.__evaluated = evaluated

    @property
    def name(self) -> str:
        return self.__name

    @property
    def passed(self) -> bool:
        return self.__expected == self.__evaluated

    @property
    def failed(self) -> bool:
        return not self.passed

    @property
    def skipped(self) -> bool:
        return self.__expected == "SKIP" and self.passed

    def failed_message(self, suite_name: str, case_number: int, case_name: str) -> str:
        return (
            f'Rule {self.name} failed on #{case_number} "{case_name}" in {suite_name}'
        )



import re
from typing import Optional, List

from cfn_guard_test.case import CfnGuardTestCase, ErrorTestCase
from cfn_guard_test.suites import CfnGuardTestSuites
from cfn_guard_test.suite import CfnGuardTestSuite
from cfn_guard_test.rule import CfnGuardRule


class CfnGuardReader:
    """
    Understands how to read the output of cfn-guard.
    """

    __suites: CfnGuardTestSuites
    __rule_name: str
    __duration: float

    def __init__(
        self, suites: CfnGuardTestSuites, rule_name: str, report: bytes, duration: float
    ):
        self.__valid_report = False
        self.__suites = suites
        self.__rule_name = rule_name
        self.__duration = duration

        self.__suite = CfnGuardTestSuite(
            name=self.__rule_name, duration=self.__duration
        )
        self.__suites.add_test_suite(self.__suite)

        sections = report.decode("utf-8").split("\n\n")
        list(map(lambda section: self.__parse_section(section), sections))

        if not self.__valid_report:
            case = ErrorTestCase(
                message=f"Unable to read the output when testing '{rule_name}'!",
                details=report.decode("utf-8"),
            )
            self.__suite.add_test_case(case)

    def __parse_section(self, test_case: str) -> None:
        if not test_case.strip():
            return None

        case_number = self.__get_case_number(test_case)
        case_name = self.__get_case_name(test_case)

        if case_number and case_name:
            self.__valid_report = True
            case = CfnGuardTestCase(name=case_name, number=case_number)
            list(map(case.add_rule, self.__get_rule_results(test_case)))

            self.__suite.add_test_case(case)

    @staticmethod
    def __get_case_number(test_case) -> Optional[int]:
        match = re.search(r"Test Case #([0-9]+)", test_case, re.MULTILINE)
        return int(match.group(1)) if match else None

    @staticmethod
    def __get_case_name(test_case: str) -> Optional[str]:
        match = re.search(r"Name: \"(.*)\"", test_case, re.MULTILINE)
        return match.group(1) if match else None

    @staticmethod
    def __get_rule_results(test_case) -> List[CfnGuardRule]:
        results = []
        matches = re.finditer(
            r"    (.*): Expected = ([A-Z]+)(, Evaluated = (\[|)([A-Z]+)(\]|)|)",
            test_case,
            re.MULTILINE,
        )

        for matchNum, match in enumerate(matches, start=1):
            rule = match.groups(matchNum)

            evaluated = rule[4] if rule[2] else rule[1]

            results.append(CfnGuardRule(str(rule[0]), str(rule[1]), str(evaluated)))

        return results



import builtins
from typing import List
import subprocess
import os
import pytest

from unittest.mock import patch, MagicMock, mock_open, call
from click.testing import CliRunner
from cfn_guard_test import main


@pytest.mark.parametrize(
    "payload, parameters, rule_exists, test_file, rule_file, expected_exit_code",
    [
        # Test 0
        ("skipped_rule.txt", [], True, "rules/iam_tests.yaml", "rules/iam.guard", 0),
        # Test 1
        ("skipped_rule.txt", [], False, "rules/iam_tests.yaml", "rules/iam.guard", 0),
        # Test 2
        (
            "skipped_rule.txt",
            ["--cfn-guard-path", "cfn-guard-fake"],
            False,
            "rules/iam_tests.yaml",
            "rules/iam.guard",
            1,
        ),
        # Test 3
        (
            "skipped_rule.txt",
            ["--cfn-guard-path", "cfn-guard"],
            True,
            "rules/iam_tests.yaml",
            "rules/iam.guard",
            0,
        ),
        # Test 4
        (
            "skipped_rule.txt",
            ["--rules-path", "my-rules"],
            True,
            "my-rules/iam_tests.yaml",
            "my-rules/iam.guard",
            0,
        ),
        # Test 5
        (
            "skipped_rule.txt",
            ["--rules-path", "my-rules", "--test-path", "my-rules"],
            True,
            "my-rules/iam_tests.yaml",
            "my-rules/iam.guard",
            0,
        ),
        # Test 6
        (
            "skipped_rule.txt",
            ["--rules-path", "my-rules", "--test-path", "my-tests"],
            True,
            "my-tests/iam_tests.yaml",
            "my-rules/iam.guard",
            0,
        ),
        # Test 7
        (
            "passing_rule.txt",
            [],
            True,
            "rules/iam_tests.yaml",
            "rules/iam.guard",
            0,
        ),
        # Test 8
        (
            "passing_rule.txt",
            ["--verbose"],
            True,
            "rules/iam_tests.yaml",
            "rules/iam.guard",
            0,
        ),
        # Test 9
        (
            "failed_rule.txt",
            [],
            True,
            "rules/iam_tests.yaml",
            "rules/iam.guard",
            1,
        ),
        # Test 10
        ("empty.txt", [], True, "rules/iam_tests.yaml", "rules/iam.guard", 1),
        # Test 11
        (
            "invalid_case_name.txt",
            [],
            True,
            "rules/iam_tests.yaml",
            "rules/iam.guard",
            1,
        ),
        # Test 12
        (
            "invalid_case_number.txt",
            [],
            True,
            "rules/iam_tests.yaml",
            "rules/iam.guard",
            1,
        ),
        # Test 13
        (
            "invalid_case_number_and_name.txt",
            [],
            True,
            "rules/iam_tests.yaml",
            "rules/iam.guard",
            1,
        ),
    ],
)
@patch("cfn_guard_test.runner.subprocess.run")
@patch("cfn_guard_test.runner.glob.glob")
@patch("cfn_guard_test.runner.os.path.isfile")
def test_invoke(
    mock_isfile: MagicMock,
    mock_glob: MagicMock,
    mock_run: MagicMock,
    payload_path: str,
    payload: str,
    parameters: List,
    rule_exists: bool,
    test_file: str,
    rule_file: str,
    expected_exit_code: int,
) -> None:
    mock_glob.return_value = [test_file]
    mock_isfile.return_value = rule_exists

    with open(f"{payload_path}/{payload}", "rb") as fh:
        mock_stdout = MagicMock()
        mock_stdout.stdout = fh.read()
        mock_run.return_value = mock_stdout

    runner = CliRunner()
    result = runner.invoke(main, parameters)
    assert result.exit_code == expected_exit_code

    if expected_exit_code == 0:
        mock_isfile.assert_called_with(rule_file)
        name = os.path.basename(test_file).split("_tests.yaml")[0]
        mock_glob.assert_called_with(test_file.replace(name, "*"))

        if rule_exists:
            mock_run.assert_called_with(
                [
                    "cfn-guard",
                    "test",
                    "--rules-file",
                    rule_file,
                    "--test-data",
                    test_file,
                ],
                stdout=subprocess.PIPE,
            )


@patch("cfn_guard_test.runner.subprocess.run")
@patch("cfn_guard_test.runner.glob.glob")
@patch("cfn_guard_test.runner.os.path.isfile")
def test_invoke_cfn_guard_failure(
    mock_isfile: MagicMock, mock_glob: MagicMock, mock_run: MagicMock
) -> None:
    mock_glob.return_value = ["rules/iam_tests.yaml"]
    mock_isfile.return_value = True

    mock_stdout = MagicMock()
    mock_stdout.stdout = bytes(
        "Error processing Parser Error when parsing Unable to process data in file...",
        "utf-8",
    )
    mock_run.return_value = mock_stdout

    runner = CliRunner()
    result = runner.invoke(main, [])
    assert result.exit_code == 1
    assert "Unable to read the output when testing 'rules/iam.guard'!" in result.stdout
    assert (
        "Error processing Parser Error when parsing Unable to process data in file..."
        in result.stdout
    )


@pytest.mark.parametrize(
    "payload, expected_exit_code",
    [
        ("passing_rule", 0),
        ("passing_rule_older_version", 0),
        ("2_passing_rules", 0),
        ("empty", 1),
        ("failed_rule", 1),
        ("failed_rule_older_version", 1),
        ("invalid_case_name", 1),
        ("invalid_case_number", 1),
        ("invalid_case_number_and_name", 1),
        ("passing_rule", 0),
        ("skipped_rule", 0),
    ],
)
@patch("cfn_guard_test.runner.subprocess.run")
@patch("cfn_guard_test.runner.glob.glob")
@patch("cfn_guard_test.runner.os.path.isfile")
@patch("cfn_guard_test.runner.time")
def test_junit(
    mock_time: MagicMock,
    mock_isfile: MagicMock,
    mock_glob: MagicMock,
    mock_run: MagicMock,
    payload_path: str,
    payload: str,
    expected_exit_code: int,
) -> None:
    mock_glob.return_value = ["rules/iam_tests.yaml"]
    mock_isfile.return_value = True
    mock_time.time.side_effect = [0.0, 0.5]

    with open(f"{payload_path}/{payload}.txt", "rb") as fh:
        mock_stdout = MagicMock()
        mock_stdout.stdout = fh.read()
        mock_run.return_value = mock_stdout

    mock_write = mock_open()
    with patch.object(builtins, "open", mock_write, create=True):
        runner = CliRunner()
        result = runner.invoke(main, ["--junit-path", "my-report-path/cfn-guard.xml"])
        assert result.exit_code == expected_exit_code

    mock_write.assert_called_with("my-report-path/cfn-guard.xml", "w")

    with open(f"{payload_path}/{payload}.xml", "rb") as fh:
        mock_write.return_value.write.assert_called_once_with(fh.read().decode("utf-8"))


def test_junit_non_writable_path() -> None:
    runner = CliRunner()
    result = runner.invoke(main, ["--junit-path", "my-report-path/cfn-guard.xml"])
    assert result.exit_code == 1


def test_junit_non_writable_file() -> None:
    mock_write = mock_open()

    mock_write.return_value.writable.return_value = False

    with patch.object(builtins, "open", mock_write, create=True):
        runner = CliRunner()
        result = runner.invoke(main, ["--junit-path", "my-report-path/cfn-guard.xml"])
        assert result.exit_code == 1



import setuptools
import toml

with open("README.md") as fp:
    long_description = fp.read()

with open("pyproject.toml") as f:
    pyproject = toml.load(f)
    project = pyproject["tool"]["poetry"]


setuptools.setup(
    name=project.get("name"),
    version=project.get("version"),
    description=project.get("description"),
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Joris Conijn",
    author_email="joris@conijnonline.nl",
    url="https://github.com/Nr18/cfn-guard-test",
    package_dir={"": "."},
    packages=setuptools.find_packages(where="."),
    install_requires=[
        "click==8.1.3",
        "junit-xml==1.9",
    ],
    entry_points={
        "console_scripts": [
            "cfn-guard-test = cfn_guard_test:main",
        ],
    },
    python_requires=">=3.8",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Code Generators",
        "Topic :: Utilities",
        "Typing :: Typed",
    ],
)
